# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

"""Entrypoint for the "wwtdatatool" command-line interface.

"""
import argparse
import os.path
import sys


# General CLI utilities

def die(msg):
    print('error:', msg, file=sys.stderr)
    sys.exit(1)

def warn(msg):
    print('warning:', msg, file=sys.stderr)


# "cabinet" subcommand

def cabinet_getparser(parser):
    subparsers = parser.add_subparsers(dest='cabinet_command')

    p = subparsers.add_parser('list')
    p.add_argument(
        'path',
        metavar = 'PATH',
        help = 'The path to a cabinet file.',
    )

    p = subparsers.add_parser('pack')
    p.add_argument(
        'cab_path',
        metavar = 'PATH',
        help = 'The path of the cabinet file to create.',
    )
    p.add_argument(
        'input_paths',
        nargs = '+',
        metavar = 'PATHS',
        help = 'Paths to files to put into the cabinet.',
    )

    p = subparsers.add_parser('unpack')
    p.add_argument(
        'path',
        metavar = 'PATH',
        help = 'The path to a cabinet file.',
    )


def cabinet_list(settings):
    from .filecabinet import FileCabinetReader

    with open(settings.path, 'rb') as f:
        reader =  FileCabinetReader(f)

        for fn in reader.filenames():
            print(fn)


def cabinet_pack(settings):
    from .filecabinet import FileCabinetWriter
    import os.path

    writer = FileCabinetWriter()

    for fn in settings.input_paths:
        with open(fn, 'rb') as f:
            data = f.read()

        # TODO: smarter splitting
        pieces = fn.split(os.path.sep)

        for p in pieces:
            if p in ('.', '..', ''):
                die(f'illegal input path "{fn}": must be relative with no ".", ".." components')

        writer.add_file_with_data('\\'.join(pieces), data)

    with open(settings.cab_path, 'wb') as f_out:
        writer.emit(f_out)


def cabinet_unpack(settings):
    from .filecabinet import FileCabinetReader
    from os import makedirs
    from os.path import join

    with open(settings.path, 'rb') as f_in:
        reader =  FileCabinetReader(f_in)

        for fn in reader.filenames():
            data = reader.read_file(fn)
            pieces = fn.split('\\')  # paths are Windows-style

            # At least the MakeDataCabinetFile tool creates a file whose
            # paths all begin with \. We are not gonna treat those as
            # absolute paths or anything like that.
            if not len(pieces[0]):
                pieces = pieces[1:]

            if len(pieces) > 1:
                makedirs(join(*pieces[:-1]), exist_ok=True)

            with open(join(*pieces), 'wb') as f_out:
                f_out.write(data)


def cabinet_impl(settings):
    if settings.cabinet_command is None:
        print('Run the "cabinet" command with `--help` for help on its subcommands')
        return

    if settings.cabinet_command == 'list':
        return cabinet_list(settings)
    elif settings.cabinet_command == 'pack':
        return cabinet_pack(settings)
    elif settings.cabinet_command == 'unpack':
        return cabinet_unpack(settings)
    else:
        die('unrecognized "cabinet" subcommand ' + settings.cabinet_command)


# "serve" subcommand

def serve_getparser(parser):
    parser.add_argument(
        '--port',
        '-p',
        metavar = 'PORT',
        type = int,
        default = 8080,
        help = 'The port on which to listen for connections.'
    )
    parser.add_argument(
        'root_dir',
        metavar = 'PATH',
        default = '.',
        help = 'The path to the base directory of the server.',
    )


def serve_impl(settings):
    from .server import run_server
    run_server(settings)


# "tree" subcommand

def tree_getparser(parser):
    subparsers = parser.add_subparsers(dest='tree_command')

    p = subparsers.add_parser('fetch')
    p.add_argument(
        'root_url',
        metavar = 'URL',
        help = 'The URL of the initial WTML file to download.',
    )

    p = subparsers.add_parser('print-dem-urls')
    p = subparsers.add_parser('print-image-urls')
    p = subparsers.add_parser('summarize')


def tree_impl(settings):
    if settings.tree_command is None:
        print('Run the "tree" command with `--help` for help on its subcommands')
        return

    if settings.tree_command == 'fetch':
        return tree_fetch(settings)
    elif settings.tree_command == 'print-dem-urls':
        return tree_print_dem_urls(settings)
    elif settings.tree_command == 'print-image-urls':
        return tree_print_image_urls(settings)
    elif settings.tree_command == 'summarize':
        return tree_summarize(settings)
    else:
        die('unrecognized "tree" subcommand ' + settings.tree_command)


def tree_fetch(settings):
    from .folder import fetch_folder_tree

    def on_fetch(url):
        print('Fetching', url, '...')

    fetch_folder_tree(settings.root_url, '.', on_fetch)


def tree_print_dem_urls(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    done_urls = set()

    for treepath, item in walk_cached_folder_tree('.'):
        imgset = None

        if isinstance(item, ImageSet):
            imgset = item
        elif isinstance(item, Place):
            imgset = item.as_imageset()

        if imgset is None:
            continue

        if not imgset.dem_url or imgset.dem_url in done_urls:
            continue

        done_urls.add(imgset.dem_url)
        print(imgset.dem_url, imgset.name)


def tree_print_image_urls(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    done_urls = set()

    for treepath, item in walk_cached_folder_tree('.'):
        imgset = None

        if isinstance(item, ImageSet):
            imgset = item
        elif isinstance(item, Place):
            imgset = item.as_imageset()

        if imgset is None:
            continue

        for url, tag in zip((imgset.url, imgset.alt_url), ('', ' (alt)')):
            if not url or url in done_urls:
                continue

            done_urls.add(url)
            print(url, imgset.name + tag)


def tree_summarize(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    for treepath, item in walk_cached_folder_tree('.'):
        pfx = '  ' * len(treepath)

        if isinstance(item, Folder):
            print(pfx + 'Folder', item.name)
        elif isinstance(item, ImageSet):
            index = treepath[-1]
            print(f'{pfx}{index:03d}', 'ImageSet:', item.name, '@', item.url)
        elif isinstance(item, Place):
            maybe_imgset = item.as_imageset()
            if maybe_imgset is not None:
                index = treepath[-1]
                print(f'{pfx}{index:03d}', 'Place+ImgSet:', item.name, '@', maybe_imgset.url)


# "wtml" subcommand

def wtml_getparser(parser):
    subparsers = parser.add_subparsers(dest='wtml_command')

    p = subparsers.add_parser('merge')
    p.add_argument(
        '--merged-name',
        default = 'Folder',
        help = 'The name to give to the merged folder.',
    )
    p.add_argument(
        '--merged-thumb-url',
        default = '',
        help = 'The thumbnail URL to give to the merged folder.',
    )
    p.add_argument(
        'in_paths',
        nargs = '+',
        metavar = 'IN-WTML-PATH',
        help = 'The path to the input WTML files.',
    )
    p.add_argument(
        'out_path',
        metavar = 'OUT-WTML-PATH',
        help = 'The path to the output WTML file.',
    )

    p = subparsers.add_parser('rewrite-urls')
    p.add_argument(
        'in_path',
        metavar = 'INPUT-WTML',
        help = 'The path to the input WTML file.',
    )
    p.add_argument(
        'baseurl',
        metavar = 'BASE-URL',
        help = 'The new base URL to use in the file\'s contents',
    )
    p.add_argument(
        'out_path',
        metavar = 'OUTPUT-WTML',
        help = 'The path of the rewritten, output WTML file.',
    )


def wtml_impl(settings):
    if settings.wtml_command is None:
        print('Run the "wtml" command with `--help` for help on its subcommands')
        return

    if settings.wtml_command == 'merge':
        return wtml_merge(settings)
    elif settings.wtml_command == 'rewrite-urls':
        return wtml_rewrite_urls(settings)
    else:
        die('unrecognized "wtml" subcommand ' + settings.wtml_command)


def wtml_merge(settings):
    from urllib.parse import urljoin, urlsplit
    from .folder import Folder

    out_folder = Folder()
    out_folder.name = settings.merged_name
    out_folder.thumbnail = settings.merged_thumb_url

    rel_base = os.path.dirname(settings.out_path)

    for path in settings.in_paths:
        in_folder = Folder.from_file(path)
        cur_base_url = path.replace(os.path.sep, '/')

        def mutator(url):
            if not url:
                return url
            if urlsplit(url).netloc:
                return url  # this URL is absolute

            # Resolve this relative URL, using the path of the source WTML
            # as the basis.
            url = urljoin(cur_base_url, url)

            # Now go back to filesystem-path land, so that we can use relpath to
            # compute the new path relative to the merged folder file.
            rel = os.path.relpath(url.replace('/', os.path.sep), rel_base)

            # Finally, re-express that as a URL
            return rel.replace(os.path.sep, '/')

        in_folder.mutate_urls(mutator)
        out_folder.children += in_folder.children

    with open(settings.out_path, 'wt', encoding='utf8') as f_out:
        out_folder.write_xml(f_out)


def wtml_rewrite_urls(settings):
    from .folder import Folder, make_absolutizing_url_mutator

    f = Folder.from_file(settings.in_path)
    f.mutate_urls(make_absolutizing_url_mutator(settings.baseurl))

    with open(settings.out_path, 'wt', encoding='utf8') as f_out:
        f.write_xml(f_out)


# The CLI driver:

def entrypoint(args=None):
    """The entrypoint for the \"wwtdatatool\" command-line interface.

    Parameters
    ----------
    args : iterable of str, or None (the default)
      The arguments on the command line. The first argument should be
      a subcommand name or global option; there is no ``argv[0]``
      parameter.

    """
    # Set up the subcommands from globals()

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    commands = set()

    for py_name, value in globals().items():
        if py_name.endswith('_getparser'):
            cmd_name = py_name[:-10].replace('_', '-')
            subparser = subparsers.add_parser(cmd_name)
            value(subparser)
            commands.add(cmd_name)

    # What did we get?

    settings = parser.parse_args(args)

    if settings.subcommand is None:
        print('Run me with --help for help. Allowed subcommands are:')
        print()
        for cmd in sorted(commands):
            print('   ', cmd)
        return

    py_name = settings.subcommand.replace('-', '_')

    impl = globals().get(py_name + '_impl')
    if impl is None:
        die('no such subcommand "{}"'.format(settings.subcommand))

    # OK to go!

    impl(settings)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

__all__ = '''
assert_xml_trees_equal
test_path
'''.split()

import os.path


TESTS_DIR = os.path.dirname(os.path.abspath(__file__))

def test_path(*pieces):
    return os.path.join(TESTS_DIR, *pieces)


def _assert_xml_trees_equal(path, e1, e2, care_text_tags):
    "Derived from https://stackoverflow.com/a/24349916/3760486"

    assert e1.tag == e2.tag, \
        'at XML path {0}, tags {1} and {2} differed'.format(path, e1.tag, e2.tag)

    # We only sometimes care about this; often it's just whitespace
    if e1.tag in care_text_tags:
        assert e1.text == e2.text, \
            'at XML path {0}, texts {1!r} and {2!r} differed'.format(path, e1.text, e2.text)

    # We never care about this, right?
    #assert e1.tail == e2.tail, \
    #    'at XML path {0}, tails {1!r} and {2!r} differed'.format(path, e1.tail, e2.tail)

    assert e1.attrib == e2.attrib, \
        'at XML path {0}, attributes {1!r} and {2!r} differed'.format(path, e1.attrib, e2.attrib)
    assert len(e1) == len(e2), \
        'at XML path {0}, number of children {1} and {2} differed'.format(path, len(e1), len(e2))

    subpath = '{0}>{1}'.format(path, e1.tag)

    for c1, c2 in zip (e1, e2):
        _assert_xml_trees_equal(subpath, c1, c2, care_text_tags)


def assert_xml_trees_equal(e1, e2, care_text_tags=()):
    _assert_xml_trees_equal('(root)', e1, e2, care_text_tags)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os.path
import shutil
import tempfile
import pytest

from .. import cli, folder
from . import test_path


@pytest.fixture
def work_in_tempdir():
    prev_dir = os.getcwd()
    d = tempfile.mkdtemp()
    os.chdir(d)
    yield d
    # Windows can't remove the temp tree unless we chdir out of it.
    os.chdir(prev_dir)
    shutil.rmtree(d)


def test_merge_cli(work_in_tempdir):
    os.mkdir('subdir1')
    os.mkdir('subdir2')

    source = folder.Folder.from_file(test_path('test1_rel.wtml'))
    test1_path = os.path.join('subdir1', 'test1_rel.wtml')
    with open(test1_path, 'w') as f:
        source.write_xml(f)

    merge_path = os.path.join('subdir2', 'merged.wtml')

    cli.entrypoint(['wtml', 'merge', test1_path, merge_path])

    merged = folder.Folder.from_file(merge_path)

    assert len(merged.children) == 1
    assert merged.children[0].thumbnail == '../subdir1/thumb.jpg'
    assert merged.children[0].foreground_image_set.credits_url == 'https://www.spacetelescope.org/images/heic0707a/'
    assert merged.children[0].foreground_image_set.url == '../subdir1/L{1}X{2}Y{3}.png'



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import pytest
from xml.etree import ElementTree as etree

from . import assert_xml_trees_equal
from .. import enums, imageset, place


def test_basic_xml():
    expected_str = '''
<Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
       Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
       DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
       Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
</Place>
'''
    expected_xml = etree.fromstring(expected_str)
    pl = place.Place()
    observed_xml = pl.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)


def test_xmeta():
    expected_str = '''
<Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
       Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
       DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
       Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0" XExtra1="hello"
       XExtra2="1">
</Place>
'''
    expected_xml = etree.fromstring(expected_str)
    pl = place.Place()
    pl.xmeta.Extra1 = 'hello'
    pl.xmeta.Extra2 = 1
    observed_xml = pl.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)


def test_nesting():
    expected_str = '''
<Place MSRCommunityId="0" MSRComponentId="0" Permission="0"
       Angle="0.0" AngularSize="0.0" DataSetType="Earth" Dec="0.0" Distance="0.0"
       DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0" Magnitude="0.0"
       Opacity="100.0" RA="0.0" Rotation="0.0" ZoomLevel="0.0">
  <BackgroundImageSet>
    <ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
              BandPass="Visible" BaseDegreesPerTile="0.0" BaseTileLevel="0"
              BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False"
              FileType=".png" Generic="False" MeanRadius="0.0" OffsetX="0.0" OffsetY="0.0" Projection="SkyImage"
              Rotation="0.0" Sparse="True" StockSet="False" TileLevels="0"
              Url="http://example.com/background" WidthFactor="2">
    </ImageSet>
  </BackgroundImageSet>
  <ForegroundImageSet>
    <ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
              BandPass="Visible" BaseDegreesPerTile="0.0" BaseTileLevel="0"
              BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False"
              FileType=".png" Generic="False" MeanRadius="0.0" OffsetX="0.0" OffsetY="0.0" Projection="SkyImage"
              Rotation="0.0" Sparse="True" StockSet="False" TileLevels="0"
              Url="http://example.com/foreground" WidthFactor="2">
    </ImageSet>
  </ForegroundImageSet>
  <ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
            BandPass="Visible" BaseDegreesPerTile="0.0" BaseTileLevel="0"
            BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False"
            FileType=".png" Generic="False" MeanRadius="0.0" OffsetX="0.0" OffsetY="0.0" Projection="SkyImage"
            Rotation="0.0" Sparse="True" StockSet="False" TileLevels="0"
            Url="http://example.com/unspecified" WidthFactor="2">
  </ImageSet>
</Place>
'''
    expected_xml = etree.fromstring(expected_str)
    pl = place.Place()
    pl.image_set = imageset.ImageSet()
    pl.image_set.url = 'http://example.com/unspecified'
    pl.foreground_image_set = imageset.ImageSet()
    pl.foreground_image_set.url = 'http://example.com/foreground'
    pl.background_image_set = imageset.ImageSet()
    pl.background_image_set.url = 'http://example.com/background'
    observed_xml = pl.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os.path
import shutil
import tempfile
import pytest

from .. import cli, filecabinet


@pytest.fixture
def tempdir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


def test_cli(tempdir):
    "Simple smoke test to see if it runs at all."

    prev_dir = os.getcwd()

    try:
        os.chdir(tempdir)
        cli.entrypoint(['tree', 'fetch',
            'https://web.wwtassets.org/engine/assets/builtin-image-sets.wtml'])
        cli.entrypoint(['tree', 'summarize'])
        cli.entrypoint(['tree', 'print-image-urls'])
        cli.entrypoint(['tree', 'print-dem-urls'])
    finally:
        # Windows can't remove the temp tree unless we chdir out of it.
        os.chdir(prev_dir)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

__all__ = '''
indent_xml
LockedDownTraits
LockedXmlTraits
MetaLockedDownTraits
stringify_xml_doc
write_xml_doc
XmlSer
'''.split()

from abc import ABCMeta
from argparse import Namespace
from enum import Enum
from traitlets import (Bool, Float, HasTraits, Instance, Int, List, MetaHasTraits,
                       TraitType, Unicode, Union, UseEnum)
from xml.etree import ElementTree as etree


class MetaLockedDownTraits(ABCMeta, MetaHasTraits):
    """A metaclass to help with the LockedDownTraits class. When a class using
    this metaclass is created, a frozenset of all of the traitlets attached to
    the class definition (including those defined by its base classes) is
    created for fast checks during __setattr__ calls. The other setup done by
    traitlet's MetaHasTraits metaclass is also performed.

    """
    def __init__(cls, name, bases, classdict):
        super(MetaLockedDownTraits, cls).__init__(name, bases, classdict)

        # This list hardcode attributes that are manipulated by the traitlets
        # machinery.
        settable_attr_names = set((
            '_cross_validation_lock',
            '_trait_notifiers',
            '_trait_validators',
            '_trait_values',
            'notify_change',
        ))

        for c in cls.mro():
            for attr_name, attr_value in c.__dict__.items():
                if isinstance(attr_value, TraitType):
                    settable_attr_names.add(attr_name)

        cls._settable_attr_names = frozenset(settable_attr_names)


class LockedDownTraits(HasTraits, metaclass=MetaLockedDownTraits):
    """A base class for HasTraits objects where we do not allow callers to add
    additional attributes to class instances. We do not allow them to delete
    attributes either.

    The general motivation is that there are some kinds of classes that
    contain a pretty fixed set of fields where it's much more likely that
    someone is going to make a typo in a variable assignment than it will be
    legitimately useful for them to add new attributes.

    There's an escape hatch, however, the ``rmeta`` attribute (for "runtime
    metadata") is a namespace object to which one can attach whichever
    metadata one wants.

    """
    rmeta = Instance(
        Namespace,
        args = (),
        help = 'Runtime metadata - a namespace object for attaching arbitrary extra information',
    )

    def __setattr__(self, name, value):
        if name not in self._settable_attr_names:
            raise AttributeError(f'not allowed to set attribute {name!r} on this instance (typo?)')
        super(LockedDownTraits, self).__setattr__(name, value)

    def __delattr__(self, name):
        if name != 'notify_change':  # sigh, traitlets
            raise AttributeError(f'not allowed to delete attribute {name!r} from this instance')
        super(LockedDownTraits, self).__delattr__(name)

    def __repr__(self):
        # This gets a bit beyond the nominal bailiwick of this class, but it's
        # pretty universally useful.
        s = ', '.join('{}={!r}'.format(a, getattr(self, a)) for a in self.trait_names())
        return f'{self.__class__.__name__}({s})'


class XmlSer(Enum):
    """Ways that a traitlet can get serialized to XML in this framework.

    """
    ATTRIBUTE = 'attr'
    TEXT_ELEM = 'text_elem'
    INNER = 'inner'
    WRAPPED_INNER = 'wrapped_inner'
    NS_TO_ATTR = 'ns_to_attr'
    INNER_LIST = 'inner_list'
    WRAPPED_INNER_LIST = 'wrapped_inner_list'

    @classmethod
    def attr(cls, attr):
        return (cls.ATTRIBUTE, attr)

    @classmethod
    def text_elem(cls, tag):
        return (cls.TEXT_ELEM, tag)

    @classmethod
    def inner(cls, tag):
        # todo: I'd like to avoid the `tag` argument, but right now we need it
        # to search for a pre-existing element when applying to an existing
        # XML tree.
        return (cls.INNER, tag)

    @classmethod
    def wrapped_inner(cls, tag):
        return (cls.WRAPPED_INNER, tag)

    @classmethod
    def ns_to_attr(cls, prefix):
        return (cls.NS_TO_ATTR, prefix)

    @classmethod
    def inner_list(cls):
        return (cls.INNER_LIST,)

    @classmethod
    def wrapped_inner_list(cls, tag):
        return (cls.WRAPPED_INNER_LIST, tag)


def _stringify_trait(trait_spec, value):
    if isinstance(trait_spec, UseEnum):
        return str(value.value)

    return str(value)


def _parse_trait(trait_spec, text):
    if isinstance(trait_spec, Float):
        return float(text)

    if isinstance(trait_spec, Int):
        return int(text)

    if isinstance(trait_spec, Unicode):
        return text

    if isinstance(trait_spec, Bool):
        try:
            return bool(int(text))
        except ValueError:
            pass

        if text.lower() == 'false':
            return False

        if text.lower() == 'true':
            return True

        raise ValueError(f'cannot interpret text {text!r} as a boolean')

    if isinstance(trait_spec, UseEnum):
        if text is None:
            text = ''
        return trait_spec.enum_class.from_text(text)

    raise ValueError(f'internal error: unimplemented parse for trait type {trait_spec}')


class LockedXmlTraits(LockedDownTraits):
    """A base class for LockedDownTraits objects that can also be serialized to
    and from XML.

    This is not a fully general XML serialization framework; it is tuned to
    the needs of the WWT XML serialization formats.

    """
    def _tag_name(self):
        """Return the XML tag name associated with the serialization of this object."""
        raise NotImplementedError()

    def _check_elem_is_this_type(self, elem):
        """Return whether the XML element represents an instance of this class.

        By default, this function returns True if the elem tag name matches
        this classes ``_tag_name()``. This filter can be overridden. This
        function is used by ``_maybe_from_xml()`` when deserializing
        containers with contents that have flexible types.

        """

        return elem.tag == self._tag_name()

    @classmethod
    def _maybe_from_xml(cls, elem):
        """Possibly deserialize an instance of this class from an XML fragment.

        Parameters
        ----------
        elem : :class:`xml.etree.ElementTree.Element`
          An XML element.

        Returns
        -------
        If the XML element appears to contain a serialized version of this class,
        returns a new instance, initialized with data from the XML. Otherwise,
        returns None.

        Notes
        -----
        This method is architected to enable deserialization of
        :class:`wwt_data_formats.folder.Folder` items. It also helps with the
        implementation of some of the "inner" XML deserialization
        implementations.

        """
        inst = cls()

        if not inst._check_elem_is_this_type(elem):
            return None

        for tname, tspec in inst.traits(xml = lambda a: a is not None).items():
            xml_spec, *xml_data = tspec.metadata['xml']
            value = None

            if xml_spec == XmlSer.ATTRIBUTE:
                text = elem.attrib.get(xml_data[0])
                if text is not None:
                    value = _parse_trait(tspec, text)
            elif xml_spec == XmlSer.TEXT_ELEM:
                sub = elem.find(xml_data[0])
                if sub is not None:
                    value = _parse_trait(tspec, sub.text)
            elif xml_spec == XmlSer.INNER:
                if not isinstance(tspec, Instance):
                    raise RuntimeError('an XML element serialized as INNER must be of Instance type')

                for sub in elem:
                    value = tspec.klass._maybe_from_xml(sub)
                    if value is not None:
                        break
            elif xml_spec == XmlSer.WRAPPED_INNER:
                if not isinstance(tspec, Instance):
                    raise RuntimeError('an XML element serialized as WRAPPED_INNER must be of Instance type')

                wrapper = elem.find(xml_data[0])
                if wrapper is not None:
                    for sub in wrapper:
                        value = tspec.klass._maybe_from_xml(sub)
                        if value is not None:
                            break
            elif xml_spec == XmlSer.NS_TO_ATTR:
                if not isinstance(tspec, Instance):
                    raise RuntimeError('an XML element serialized as NS_TO_ATTR must be of Instance type')

                value = tspec.klass()
                for aname, avalue in elem.attrib.items():
                    if aname.startswith(xml_data[0]):
                        setattr(value, aname[len(xml_data[0]):], avalue)
            elif xml_spec == XmlSer.INNER_LIST:
                if not isinstance(tspec, List):
                    raise RuntimeError('XML elements serialized as INNER_LIST must be of List type')

                # total hackiness specific for the Folder use case, plus encapsulation breakage:
                un_spec = tspec._trait
                if not isinstance(un_spec, Union):
                    raise RuntimeError('XML elements serialized as INNER_LIST must be of List(Union) type')

                klasses = [t.klass for t in un_spec.trait_types]
                # end hackiness, maybe.

                value = []

                for sub in elem:
                    for klass in klasses:
                        v = klass._maybe_from_xml(sub)
                        if v is not None:
                            value.append(v)
                            break
            elif xml_spec == XmlSer.WRAPPED_INNER_LIST:
                # Repeat the INNER_LIST hackiness
                if not isinstance(tspec, List):
                    raise RuntimeError('XML elements serialized as WRAPPED_INNER_LIST must be of List type')

                un_spec = tspec._trait
                if not isinstance(un_spec, Union):
                    raise RuntimeError('XML elements serialized as WRAPPED_INNER_LIST must be of List(Union) type')

                klasses = [t.klass for t in un_spec.trait_types]
                # end hackiness, maybe.

                value = []

                wrapper = elem.find(xml_data[0])
                if wrapper is not None:
                    for sub in wrapper:
                        for klass in klasses:
                            v = klass._maybe_from_xml(sub)
                            if v is not None:
                                value.append(v)
                                break
            else:
                raise RuntimeError(f'unhandled XML serialization mode {xml_spec}')

            if value is not None:
                setattr(inst, tname, value)

        return inst


    @classmethod
    def from_xml(cls, elem):
        """Deserialize an instance of this class from XML.

        Parameters
        ----------
        elem : xml.etree.ElementTree.Element
          An XML element serializing the object.

        Returns
        -------
        An instance of the class, initialized with data from the XML.

        """
        inst = cls._maybe_from_xml(elem)
        if inst is None:
            raise ValueError(f'expected to get a {cls} instance from <{elem.tag}>, but didn\'t')
        return inst


    @classmethod
    def from_text(cls, text):
        """Deserialize an instance of this class from XML-formatted text.

        Parameters
        ----------
        text : string
          The XML text.

        Returns
        -------
        An instance of the class, initialized with data from the XML.

        """
        elem = etree.fromstring(text)
        return cls.from_xml(elem)


    @classmethod
    def from_file(cls, path, encoding='utf-8-sig'):
        """Deserialize an instance of this class from an XML file on local disk.

        Parameters
        ----------
        path : string
            The path of the XML file.
        encoding : optional string, default "utf-8-sig"
            The encoding of the file text. The default value is basically UTF-8 but
            will ignore Windows Byte Order Markers (BOMs) if present.

        Returns
        -------
        An instance of the class, initialized with data from the XML.

        """
        with open(path, 'rt', encoding=encoding) as f:
            text = f.read()

        return cls.from_text(text)


    @classmethod
    def from_url(cls, url, session=None, **kwargs):
        """Deserialize an instance of this class from XML downloaded from the
        specified URL.

        Parameters
        ----------
        url : string
            The URL from which to download the XML.
        session : ``requests`` session or None (the default)
            The HTTP communications session to use.
        kwargs
            Extra arguments to pass to ``requests.get``.

        Returns
        -------
        An instance of the class, initialized with data from the XML.

        """
        if session is None:
            import requests
            session = requests

        resp = session.get(url, **kwargs)

        # We have to set this to get requests/Python to ignore the Unicode
        # Byte Order Marker (BOM) that is present in some WWT data files due
        # to their Windows origin, when we decode the response into text.
        resp.encoding = 'utf-8-sig'

        elem = etree.fromstring(resp.text)
        return cls.from_xml(elem)


    def _serialize_xml(self, elem):
        """Do the work of serializing this thing to XML.

        Parameters
        ----------
        elem : :class:`xml.etree.ElementTree.Element` or None
          An XML subtree that will be filled in with data values associated with this
          instance. If None, a new element will be created and populated with the
          instance data.

        Returns
        -------
        The serialized XML element. If it was not None, this is the parameter *elem*.

        Notes
        -----
        This method is intended to allow two modes of operation. It can either
        create XML trees from scratch, or it can modify existing trees. The
        latter mode makes it possible to work with input data that may have
        additional contents not defined in the WWT specification -- by
        modifying the existing tree, we can preserve those data.

        A trait whose value is None is omitted from the XML serialization.

        A trait that is represented textually (either as an attribute or a
        text element) whose textualization is empty is also ommitted from the
        XML serialization.

        """
        if elem is None:
            elem = etree.Element(self._tag_name())

        # Note that self.traits() generates the traits in alphabetical order,
        # which is helpful because that means when we serialize to XML the
        # child elements are created in a deterministic order, which makes it
        # easier to test the output.

        for tname, tspec in self.traits(xml = lambda a: a is not None).items():
            xml_spec, *xml_data = tspec.metadata['xml']
            value = getattr(self, tname)

            if value is None:
                continue

            if xml_spec == XmlSer.ATTRIBUTE:
                text = _stringify_trait(tspec, value)
                if not text:
                    continue

                elem.set(xml_data[0], text)
            elif xml_spec == XmlSer.TEXT_ELEM:
                text = _stringify_trait(tspec, value)
                if not text:
                    continue

                sub = elem.find(xml_data[0])
                if sub is None:
                    sub = etree.SubElement(elem, xml_data[0])

                sub.text = text
            elif xml_spec == XmlSer.INNER:
                cur_sub = elem.find(xml_data[0])
                new_sub = value._serialize_xml(cur_sub)
                if cur_sub is None:
                    elem.append(new_sub)
            elif xml_spec == XmlSer.WRAPPED_INNER:
                wrapper = elem.find(xml_data[0])
                if wrapper is None:
                    wrapper = etree.SubElement(elem, xml_data[0])

                if len(wrapper):
                    value._serialize_xml(wrapper[0])
                else:
                    new_sub = value._serialize_xml(None)
                    wrapper.append(new_sub)
            elif xml_spec == XmlSer.NS_TO_ATTR:
                for ns_name, ns_value in value.__dict__.items():
                    elem.set(xml_data[0] + ns_name, str(ns_value))
            elif xml_spec == XmlSer.INNER_LIST:
                # This is gross. If the list of children gets mutated, in the
                # current framework we basically have no way to know what
                # pre-existing XML elements map to our children. The best I
                # can come up with is that we try to enforce the invariant
                # that the list of children cannot get mutated.
                preexisting = bool(len(elem))

                if preexisting and len(elem) != len(value):
                    raise RuntimeError('serializing flexible list to existing XML data, '
                                       'but it looks like something changed beneath us')

                for idx, child in enumerate(value):
                    if preexisting:
                        if elem[idx].tag != child._tag_name():
                            raise RuntimeError('serializing flexible list to existing XML data, '
                                               f'but it looks like child #{i} changed')
                        child._serialize_xml(elem[idx])
                    else:
                        new_sub = child._serialize_xml(None)
                        elem.append(new_sub)
            elif xml_spec == XmlSer.WRAPPED_INNER_LIST:
                wrapper = elem.find(xml_data[0])

                if wrapper is None:
                    preexisting = False
                    wrapper = etree.SubElement(elem, xml_data[0])
                else:
                    preexisting = True

                    if len(wrapper) != len(value):
                        raise RuntimeError('serializing flexible list to existing XML data, '
                                           'but it looks like something changed beneath us')

                for idx, child in enumerate(value):
                    if preexisting:
                        if wrapper[idx].tag != child._tag_name():
                            raise RuntimeError('serializing flexible list to existing XML data, '
                                               f'but it looks like child #{i} changed')
                        child._serialize_xml(wrapper[idx])
                    else:
                        new_sub = child._serialize_xml(None)
                        wrapper.append(new_sub)
            else:
                raise RuntimeError(f'unhandled XML serialization mode {xml_spec}')

        return elem


    def apply_to_xml(self, elem):
        """Serialize the data of this object to an existing XML tree

        Parameters
        ----------
        elem : :class:`xml.etree.ElementTree.Element`
          An XML subtree that will be filled in with data values associated with this
          instance.

        Returns
        -------
        *self*

        Notes
        -----
        This method makes it possible to work with input data that may have
        additional contents not defined in the WWT specification -- by
        modifying the existing tree, we can preserve those data.

        """
        self._serialize_xml(elem)
        return self


    def to_xml(self):
        """Serialize this object to XML.

        Returns
        -------
        elem : :class:`xml.etree.ElementTree.Element`
          An XML element serializing the object.

        """
        return self._serialize_xml(None)


    def write_xml(self, dest_stream, dest_wants_bytes=False, indent=True):
        """Serialize this object to XML, writing the data to a stream.

        Parameters
        ----------
        dest_stream : writeable file-like object or None
          The destination to which the XML data will be written. If None,
          standard output is used.
        dest_wants_bytes : optional bool, default False
          Whether the destination stream expects to be fed bytes data rather
          than Unicode. If so, the XML text is encoded into UTF-8 before
          being written.
        indent : bool, default True
          Whether the returned XML text will have user-friendly indentation or not.

        Returns
        -------
        None

        """
        write_xml_doc(
            self.to_xml(),
            dest_stream = dest_stream,
            dest_wants_bytes = dest_wants_bytes,
            indent = indent,
        )


    def to_xml_string(self, indent=True):
        """Serialize this object to XML text.

        Parameters
        ----------
        indent : optional bool, default True
          Whether the returned XML text will have user-friendly indentation or not.

        Returns
        -------
        xml_text : string
          A textual serialization of the object as XML.

        """
        return stringify_xml_doc(self.to_xml(), indent=indent)


def indent_xml(elem, level=0):
    """A dumb XML indenter.

    We create XML files using xml.etree.ElementTree, which is careful about
    spacing and so by default creates ugly files with no linewraps or
    indentation. This function is copied from `ElementLib
    <http://effbot.org/zone/element-lib.htm#prettyprint>`_ and implements
    basic, sensible indentation using "tail" text.

    """
    i = "\n" + level * "  "

    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:  # intentionally updating "elem" here!
            indent_xml(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def write_xml_doc(root_element, indent=True, dest_stream=None, dest_wants_bytes=False):
    """Write out some XML elements, indenting them by default

    When *indent* is true, the default setting, this function modifies *root_element*.

    If *dest_stream* is left unspecified, ``sys.stdout`` is used.

    """
    if dest_stream is None:
        import sys
        dest_stream = sys.stdout

    if indent:
        indent_xml(root_element)

    doc = etree.ElementTree(root_element)

    # We could in principle auto-detect this with a 0-byte write(), I guess?
    if dest_wants_bytes:
        encoding = 'UTF-8'
    else:
        encoding = 'Unicode'

    doc.write(dest_stream, encoding=encoding, xml_declaration=True)


def stringify_xml_doc(root_element, indent=True):
    """Stringify some XML elements, indenting them by default.

    When *indent* is true, the default setting, this function modifies *root_element*.

    """
    from io import StringIO
    with StringIO() as dest:
        write_xml_doc(root_element, indent, dest, dest_wants_bytes=False)
        result = dest.getvalue()
    return result



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

"""Serialized information about graphical "layers" in the WWT engine.

Not implemented:

- GreatCircleRouteLayer
- GroundOverlayLayer
- KmlLayer
- Object3d
- OrbitLayer
- SpreadSheetLayer
- TimeSeriesLayer
- VoTableLayer
- WmsLayer

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
ImageSetLayer
Layer
LayerContainerReader
LayerContainerXml
'''.split()

from traitlets import Bool, Float, Instance, List, Unicode, Union

from . import LockedXmlTraits, XmlSer
from .filecabinet import FileCabinetReader
from .imageset import ImageSet


class LayerContainerReader(object):
    """A collection of layers and reference frames."""

    _reader = None
    _info = None

    def __init__(self, stream):
        self._reader = FileCabinetReader(stream)

        for fn in self._reader.filenames():
            if fn.endswith('.wwtxml') and '\\' not in fn:
                b = self._reader.read_file(fn)
                text = b.decode('utf-8-sig')
                self._info = LayerContainerXml.from_text(text)

        if self._info is None:
            raise Exception('found no ".wwtxml" file in WWTL file cabinet')


    def close(self):
        self._reader.close()
        self._reader = None


    @classmethod
    def from_file(cls, path):
        """Deserialize layers from a WWTL "file cabinet" file.

        Parameters
        ----------
        path : string
          The path of the file-cabinet file.

        Returns
        -------
        An initialized instance of the class.

        """
        f = open(path, 'rb')
        return cls(f)


    @property
    def layers(self):
        return self._info.layers


    def read_layer_file(self, layer, extension):
        """Read a data file associated with a layer in this container.

        Parameters
        ----------
        layer : :class:`Layer`
            One of the layer objects found in this container.
        extension : string
            The file extension of the associated data file, including
            a leading period. For instance, ".txt".

        Returns
        -------
        The contents of the data file as a :class:`bytes` object.

        """
        # In most examples I've seen, the layer file is in a subdirectory
        # grouped by the layer's UUID. But in one example (David Weigel
        # LMC WWTL file, June 2020), it's not in the subdirectory.

        fn = self._info.id + '\\' + layer.id + extension

        if fn in self._reader.filenames():
            return self._reader.read_file(fn)
        return self._reader.read_file(layer.id + extension)


class LayerContainerXml(LockedXmlTraits):
    """The XML serialization of the layer collection information."""

    id = Unicode('').tag(xml=XmlSer.attr('ID'))
    """A UUID-format randomized identifier for this layer collection.

    Data files associated with this layer collection that come in the WWTL
    cabinet file will be placed in a subdirectory whose name is this ID.

    """

    # ignore ReferenceFrames for now

    layers = List(
        trait = Union([
            Instance('wwt_data_formats.layers.ImageSetLayer', args=()),
        ]),
        default_value = ()
    ).tag(xml=XmlSer.wrapped_inner_list('Layers'))
    "A list of the layers stored in this container."

    def _tag_name(self):
        return 'LayerContainer'


class Layer(LockedXmlTraits):
    """Generic parent class for serializable WWT layers."""

    id = Unicode('').tag(xml=XmlSer.attr('Id'))
    """A UUID-format randomized identifier for this layer collection.

    Data files associated with this layer collection that come in the WWTL
    cabinet file will be placed in a subdirectory whose name is this ID.

    """

    layer_type = Unicode('').tag(xml=XmlSer.attr('Type'))
    "A textual representation of this layer's type."

    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    "A user-facing name for this layer."

    reference_frame = Unicode('').tag(xml=XmlSer.attr('ReferenceFrame'))
    "The name of the reference frame relative to which this layer is positioned."

    opacity = Float(1.0).tag(xml=XmlSer.attr('Opacity'))

    # Skipping: Color, StartTime, Endtime, FadeSpan, FadeType

    def _tag_name(self):
        return 'Layer'

    def _layertype_name(self):
        raise NotImplementedError()

    def _check_elem_is_this_type(self, elem):
        return elem.attrib.get('Type', '') == self._layertype_name()


class ImageSetLayer(Layer):
    extension = Unicode('').tag(xml=XmlSer.attr('Extension'))
    """The filename extension for the underlying image if this is a SkyImage
    layer. This includes the period, e.g. ".tif"."""

    override_default = Bool(False).tag(xml=XmlSer.attr('OverrideDefault'))

    image_set = Instance(ImageSet).tag(xml=XmlSer.inner('ImageSet'))

    # Skipping: ScaleType, MinValue, MaxValue

    def _layertype_name(self):
        return 'TerraViewer.ImageSetLayer'



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the .NET Foundation
# Licensed under the MIT License.

"""Some abstract base classes (ABCs).

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
UrlContainer
'''.split()

from abc import ABC, abstractmethod


class UrlContainer(ABC):
    """A data object that may contain URLs. This ABC provides a generic
    interface for mutating those URLs.

    This ABC helps establish a simple framework for walking a WWT folder structure
    and rewriting relative into absolute URLs.

    """

    @abstractmethod
    def mutate_urls(self, mutator):
        """Visit all URLs inside this container and potentially mutate them.

        Parameters
        ----------
        mutator : callable(str) -> str
            A function taking a URL string and returning a URL string.

        Notes
        -----
        For each URL inside the container, the URL is updated with the mutator's return value::

            this.some_url = mutator(this.some_url)

        If you just want to visit all of the URLs inside this container, you can write a mutator
        the returns its input unmodified.

        Implementors should not call the mutator if the URL is an optional parameter that is ``None``
        (or otherwise unset).
        """



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License

from __future__ import absolute_import, division, print_function

import os
from setuptools import setup, Extension

def get_long_desc():
    in_preamble = True
    lines = []

    with open('README.md', 'rt', encoding='utf8') as f:
        for line in f:
            if in_preamble:
                if line.startswith('<!--pypi-begin-->'):
                    in_preamble = False
            else:
                if line.startswith('<!--pypi-end-->'):
                    break
                else:
                    lines.append(line)

    lines.append('''

For more information, including installation instructions, please visit [the
project homepage].

[the project homepage]: https://wwt-data-formats.readthedocs.io/
''')
    return ''.join(lines)


setup_args = dict(
    name = 'wwt_data_formats',  # cranko project-name
    version = '0.5.0',  # cranko project-version
    description = 'Low-level interface to AAS WorldWide Telescope data formats',
    long_description = get_long_desc(),
    long_description_content_type = 'text/markdown',
    url = 'https://wwt-data-formats.readthedocs.io/',
    license = 'MIT',
    platforms = 'Linux, Mac OS X',

    author = 'AAS WorldWide Telescope Team',
    author_email = 'wwt@aas.org',

    classifiers = [
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Visualization',
    ],

    packages = [
        'wwt_data_formats',
        'wwt_data_formats.tests',
    ],
    include_package_data = True,

    entry_points = {
        'console_scripts': [
            'wwtdatatool=wwt_data_formats.cli:entrypoint',
        ],
    },

    install_requires = [
        'requests',
        'traitlets',
    ],

    extras_require = {
        'test': [
            'mock',
            'pytest-cov',
        ],
        'docs': [
            'astropy-sphinx-theme',
            'numpydoc',
            'sphinx',
            'sphinx-automodapi',
        ],
    },
)

if __name__ == '__main__':
    setup(**setup_args)
